import logging
import os
import subprocess
import sys

import ruamel.yaml

from . import Collection, Entry, TemplatingError


def edit_file_in_external_editor(filepath):
    # First we try to use $EDITOR
    try:
        editor = os.environ['EDITOR']
        subprocess.call([editor, filepath])
    except FileNotFoundError as e:
        if editor in str(e):
            logging.info(
                'Could not open file {} with $EDITOR (currently {})'.format(
                    filepath,
                    editor
                    )
                )
        else:
            logging.warning("Could not open file {}".format(filepath))
            return False
    else:
        return True
    # then we try to use sensible-editor (which should be available on
    # debian and derivatives)
    try:
        subprocess.call(['sensible-editor', filepath])
    except FileNotFoundError as e:
        if 'sensible-editor' in e.strerror:
            logging.debug(
                "Could not open file {} with editor: sensible-editor".format(
                    filepath
                    )
                )
        else:
            logging.warning("Could not open file {}".format(filepath))
            return False
    else:
        return True
    # and finally we fallback to vi, because ed is the standard editor,
    # but that would be way too cruel, and vi is also in posix
    try:
        subprocess.call(['vi', filepath])
    except FileNotFoundError as e:
        if 'vi' in e.strerror:
            logging.warning(
                "Could not open file {} with any known editor".format(filepath)
                )
            return False
        else:
            logging.warning("Could not open file {}".format(filepath))
            return False
    else:
        return True


class Command():
    help = ''

    def _main(self, args):
        self.args = args
        self.main()


class New(Command):
    arguments = [
        (['--collection', '-c'], dict(
            help='The collection to work on (default .)'
            )),
        (['--no-git'], dict(
            help="Don't add the new entry to git",
            action="store_false",
            dest='git'
            )),
        ]

    def main(self):
        collection = Collection(self.args.collection)
        new_entry = Entry(collection)
        collection.save_entries([new_entry])
        filepath = os.path.join(
            collection.itemdir,
            new_entry.fname
            )
        if edit_file_in_external_editor(filepath):
            collection.update_cache([filepath])
            if self.args.git:
                collection.git_add_files([filepath])
        saved_entry = collection.entry_from_uid(new_entry.uid)
        print(saved_entry)


class Edit(Command):
    arguments = [
        (['--collection', '-c'], dict(
            help='The collection to work on (default .)'
            )),
        (['--no-git'], dict(
            help="Don't add the new entry to git",
            action="store_false",
            dest='git'
            )),
        (['uid'], dict(
            help='uid of an entry to edit',
            )),
        ]

    def main(self):
        collection = Collection(self.args.collection)
        entries = collection.entries_from_short_uid(self.args.uid)
        if len(entries) > 1:
            return "{} is not an unique uid".format(self.args.uid)
        if not entries:
            return "Could not find an entry with uid starting with: {}".format(
                self.args.uid
                )
        entry = entries[0]
        filepath = os.path.join(
            collection.itemdir,
            entry.fname
            )
        if edit_file_in_external_editor(filepath):
            collection.update_cache([filepath])
            if self.args.git:
                collection.git_add_files([filepath])
        saved_entry = collection.entry_from_uid(entry.uid)
        print(saved_entry)


class Show(Command):
    arguments = [
        (['--collection', '-c'], dict(
            help='The collection to work on (default .)'
            )),
        (['--template', '-t'], dict(
            help='Use the specified template to display results.',
            )),
        (['uid'], dict(
            help='uid of an entry to edit',
            )),
        ]

    def main(self):
        collection = Collection(self.args.collection)
        entries = collection.entries_from_short_uid(self.args.uid)
        if len(entries) > 1:
            return "{} is not an unique uid".format(self.args.uid)
        if not entries:
            return "Could not find an entry with uid starting with: {}".format(
                self.args.uid
                )
        entry = entries[0]
        if self.args.template:
            try:
                print(entry.render(self.args.template))
            except TemplatingError as e:
                logging.error("{}".format(e))
                sys.exit(1)
        else:
            print(entry.yaml_data)


class Index(Command):
    arguments = [
        (['--collection', '-c'], dict(
            help='The collection to work on (default .)'
            )),
        (['files'], dict(
            help='List of files to index (default: everything)',
            default=None,
            nargs='*',
            )),
        ]

    def main(self):
        collection = Collection(self.args.collection)
        if self.args.files:
            files = (os.path.basename(f) for f in self.args.files)
        else:
            files = None
        indexed = collection.update_cache(fnames=files)
        print("Found and indexed {} entries".format(indexed))


class Search(Command):
    arguments = [
        (['--collection', '-c'], dict(
            help='The collection to work on (default .)'
            )),
        (['--template', '-t'], dict(
            help='Template to use when displaying results',
            )),
        (['--offset'], dict(
            type=int,
            )),
        (['--pagesize'], dict(
            type=int,
            )),
        (['--all'], dict(
            action='store_true',
            help='Return all available results'
            )),
        (['query'], dict(
            help='Xapian query to search in the collection',
            nargs='+'
            )),
        ]

    def main(self):
        # TODO: implement "searching" for everything
        if self.args.offset:
            logging.warning(
                "offset exposes an internal knob and MAY BE" +
                " REMOVED from a future release of lesana"
                )
        if self.args.pagesize:
            logging.warning(
                "pagesize exposes an internal knob and MAY BE" +
                " REMOVED from a future release of lesana"
                )
        offset = self.args.offset or 0
        pagesize = self.args.pagesize or 12
        collection = Collection(self.args.collection)
        if self.args.query == ['*']:
            results = collection.get_all_documents()
        else:
            collection.start_search(' '.join(self.args.query))
            if self.args.all:
                results = collection.get_all_search_results()
            else:
                results = collection.get_search_results(
                    offset,
                    pagesize)
        if self.args.template:
            try:
                template = collection.get_template(self.args.template)
                print(template.render(entries=results))
            except TemplatingError as e:
                logging.error("{}".format(e))
                sys.exit(1)
        else:
            for entry in results:
                print("{entry}".format(
                    entry=entry,
                    ))


class Export(Command):
    arguments = [
        (['--collection', '-c'], dict(
            help='The collection to work on (default .)'
            )),
        (['--query', '-q'], dict(
            help='Xapian query to search in the collection',
            )),
        (['destination'], dict(
            help='The collection to export entries to'
            )),
        (['template'], dict(
            help='Template to convert entries',
            )),
        ]

    def main(self):
        collection = Collection(self.args.collection)
        destination = Collection(self.args.destination)
        if not self.args.query:
            results = collection.get_all_documents()
        else:
            collection.start_search(' '.join(self.args.query))
            results = collection.get_all_search_results()
        for entry in results:
            try:
                template = collection.get_template(self.args.template)
                rendered = template.render(entry=entry, **entry.data)
            except TemplatingError as e:
                logging.error("Error converting entry: {}".format(entry))
                logging.error("{}".format(e))
                sys.exit(1)
            try:
                data = ruamel.yaml.load(rendered, ruamel.yaml.RoundTripLoader)
            except ruamel.yaml.YAMLError as e:
                logging.error("Error converting entry: {}".format(entry))
                logging.error("{}".format(e))
                sys.exit(1)
            e = Entry(destination, data=data)
            destination.save_entries([e])


class Init(Command):
    arguments = [
        (['--collection', '-c'], dict(
            help='The directory to work on (default .)',
            default='.'
            )),
        (['--no-git'], dict(
            help='Skip setting up git in this directory',
            action="store_false",
            dest='git'
            )),
        ]

    def main(self):
        Collection.init(
            self.args.collection,
            git_enabled=self.args.git,
            edit_file=edit_file_in_external_editor
            )


class Remove(Command):
    arguments = [
        (['--collection', '-c'], dict(
            help='The collection to work on (default .)',
            )),
        (['entries'], dict(
            help='List of entries to remove',
            nargs='+',
            )),
        ]

    def main(self):
        collection = Collection(self.args.collection)
        collection.remove_entries(uids=self.args.entries)



import decimal
import logging
import os
import uuid

import ruamel.yaml
import xapian
import jinja2

from pkg_resources import resource_string

try:
    import git
    git_available = True
except ImportError:
    git_available = False


class Entry(object):
    def __init__(self, collection, data={}, fname=None):
        self.collection = collection
        self.data = data or self.empty_data()
        self.fname = fname
        self.uid = self.data.get('uid', None)
        if not self.uid:
            if self.fname:
                self.uid, ext = os.path.splitext(os.path.basename(self.fname))
            else:
                self.uid = uuid.uuid4().hex
        if not self.fname:
            self.fname = self.uid + '.yaml'

    def __str__(self):
        label = self.collection.settings.get('entry_label', None)
        if label:
            t = jinja2.Template(label)
            return t.render(**self.get_data())
        else:
            return self.uid

    def get_data(self):
        d = self.data.copy()
        d['uid'] = self.uid
        d['fname'] = self.fname
        d['short_id'] = self.short_id
        return d

    def empty_data(self):
        data = ''
        for field in self.collection.settings['fields']:
            if not field.get('help', None) is None:
                data += "# {name} ({type}): {help}\n".format(**field)
            t = field['type']
            if field.get('default', None):
                data += ("{name}: '{default}'\n".format(**field))
            elif t == 'string':
                data += ("{name}: ''\n".format(**field))
            elif t == 'text':
                data += ("{name}: |\n    .\n".format(**field))
            elif t == 'integer':
                data += ("{name}: 0\n".format(**field))
            elif t == 'float':
                data += ("{name}: 0.0\n".format(**field))
            elif t == 'decimal':
                data += ("{name}: 0.00\n".format(**field))
            elif t == 'list':
                data += ("{name}: []\n".format(**field))
            else:
                data += ("{name}: \n".format(**field))
        return ruamel.yaml.load(data, Loader=ruamel.yaml.RoundTripLoader)

    @property
    def yaml_data(self):
        return ruamel.yaml.dump(self.data, Dumper=ruamel.yaml.RoundTripDumper)

    @property
    def idterm(self):
        return "Q"+self.uid

    @property
    def short_id(self):
        return self.uid[:8]

    def validate(self):
        errors = []
        valid = True
        for field in self.collection.settings['fields']:
            value = self.data.get(field['name'], None)
            if not value:
                # empty fields are always fine
                continue
            t = field['type']
            if t == 'integer':
                try:
                    int(value)
                except ValueError:
                    valid = False
                    errors.append({
                        'field': field['name'],
                        'error': 'Invalid value for integer field: {}'.format(
                            value
                            ),
                        })
            elif t == 'float':
                try:
                    float(value)
                except ValueError:
                    valid = False
                    errors.append({
                        'field': field['name'],
                        'error': 'Invalid value for float field: {}'.format(
                            value
                            ),
                        })
            elif t == 'decimal':
                try:
                    decimal.Decimal(value)
                except ValueError:
                    valid = False
                    errors.append({
                        'field': field['name'],
                        'error': 'Invalid value for decimal field: {}'.format(
                            value
                            ),
                        })
        return valid, errors

    def render(self, template, searchpath='.'):
        jtemplate = self.collection.get_template(template, searchpath)
        try:
            return jtemplate.render(entry=self)
        except jinja2.exceptions.TemplateSyntaxError as e:
            raise TemplatingError('Template Syntax Error: '+str(e))


class Collection(object):
    """
    """
    PARSER_FLAGS = (
        xapian.QueryParser.FLAG_BOOLEAN |
        xapian.QueryParser.FLAG_PHRASE |
        xapian.QueryParser.FLAG_LOVEHATE |
        xapian.QueryParser.FLAG_WILDCARD
        )

    def __init__(self, directory=None, itemdir='items'):
        self.basedir = directory or os.getcwd()
        self.itemdir = os.path.join(self.basedir, itemdir)
        try:
            with open(os.path.join(self.basedir, 'settings.yaml')) as fp:
                self.settings = ruamel.yaml.load(
                    fp,
                    ruamel.yaml.RoundTripLoader
                    )
        except FileNotFoundError:
            self.settings = ruamel.yaml.safe_load("{}")
        os.makedirs(os.path.join(self.basedir, '.lesana'), exist_ok=True)
        if 'lang' in self.settings:
            try:
                self.stemmer = xapian.Stem(self.settings['lang'])
            except xapian.InvalidArgumentError:
                logging.warning(
                    "Invalid language %s, in settings.yaml: using english.",
                    self.settings['lang']
                    )
                self.stemmer = xapian.Stem('english')
        else:
            self.stemmer = xapian.Stem('english')
        self._enquire = None
        # This selects whether to load all other yaml files with
        # safe_load or load + RoundTripLoader
        self.safe = False
        self.entry_class = Entry

    def _index_file(self, fname, cache):
        with open(os.path.join(self.itemdir, fname)) as fp:
            if self.safe:
                data = ruamel.yaml.safe_load(fp)
            else:
                data = ruamel.yaml.load(fp, ruamel.yaml.RoundTripLoader)
        entry = self.entry_class(self, data, fname)
        valid, errors = entry.validate()
        if not valid:
            logging.warning(
                "Not indexing {fname}: invalid data".format(fname=fname)
                )
            return False, errors

        doc = xapian.Document()
        self.indexer.set_document(doc)

        # Fields with prefix, for field search
        for field in self.indexed_fields:
            if field['multi']:
                values = entry.data.get(field['name'])
            else:
                values = [entry.data.get(field['name'])]
            if not values:
                values = []
            for v in values:
                try:
                    self.indexer.index_text(
                        str(v),
                        1,
                        field['prefix'])
                except ValueError as e:
                    logging.info("Not indexing empty? value {}: {}".format(
                        entry.data.get(field['name']),
                        str(e)))
        # unprefixed fields, for full text search
        for field in self.indexed_fields:
            if field.get('free_search', False):
                if field['multi']:
                    values = entry.data.get(field['name'])
                else:
                    values = [entry.data.get(field['name'])]
                if not values:
                    values = []
                for v in values:
                    try:
                        self.indexer.index_text(v)
                        self.indexer.increase_termpos()
                    except ValueError as e:
                        # probably already logged earlier
                        pass
        doc.set_data(entry.yaml_data)
        doc.add_boolean_term(entry.idterm)
        doc.add_value(0, entry.fname.encode('utf-8'))

        cache.replace_document(entry.idterm, doc)
        return True, []

    @property
    def indexed_fields(self):
        fields = []
        for field in self.settings['fields']:
            if field.get('index', '') in ['free', 'field']:
                prefix = field.get('prefix', 'X'+field['name'].upper())
                fields.append({
                    'prefix': prefix,
                    'name': field['name'],
                    'free_search': field['index'] == 'free',
                    'multi': field['type'] in ['list']
                    })
        return fields

    def update_cache(self, fnames=None):
        """
        Update the xapian db with the data in files.

        ``fnames`` is a list of *basenames* of files in ``self.itemdir``.

        If no files have been passed, add everything.

        Return the number of files that have been added to the cache.
        """
        cache = xapian.WritableDatabase(
            os.path.join(self.basedir, '.lesana/xapian'),
            xapian.DB_CREATE_OR_OPEN
            )
        self.indexer = xapian.TermGenerator()
        self.indexer.set_stemmer(self.stemmer)
        if not fnames:
            try:
                fnames = os.listdir(self.itemdir)
            except FileNotFoundError:
                logging.warning(
                    "No such file or directory: %s, not updating cache".format(
                        self.itemdir)
                    )
                return 0
        updated = 0
        for fname in fnames:
            try:
                valid, errors = self._index_file(fname, cache)
            except IOError as e:
                logging.warning("Could not load file {}: {}".format(
                    fname,
                    str(e))
                    )
            else:
                if valid:
                    updated += 1
                else:
                    logging.warning(
                        "File {fname} could not be indexed: {errors}".format(
                            fname=fname,
                            errors=errors)
                        )
        return updated

    def save_entries(self, entries=[]):
        for e in entries:
            complete_name = os.path.join(
                self.itemdir,
                e.fname
                )
            with open(complete_name, 'w') as fp:
                fp.write(e.yaml_data)

    def git_add_files(self, files=[]):
        if not git_available:
            logging.warning(
                "python3-git not available, could not initalise " +
                "the git repository.")
            return False
        if not self.settings.get('git', False):
            logging.info(
                "This collection is configured not to use git"
                )
            return False
        try:
            repo = git.Repo(self.basedir, search_parent_directories=True)
        except git.exc.InvalidGitRepositoryError as e:
            logging.warning(
                "Could not find a git repository in {}".format(
                    self.basedir))
            return False
        repo.index.add(files)
        return True

    def _get_cache(self):
        try:
            cache = xapian.Database(
                os.path.join(self.basedir, '.lesana/xapian'),
                )
        except xapian.DatabaseOpeningError:
            logging.info("No database found, indexing entries.")
            self.update_cache()
            cache = xapian.Database(
                os.path.join(self.basedir, '.lesana/xapian'),
                )
        return cache

    def start_search(self, querystring):
        """
        Prepare a search for querystring.
        """
        cache = self._get_cache()
        queryparser = xapian.QueryParser()
        queryparser.set_stemmer(self.stemmer)
        queryparser.set_database(cache)

        for field in self.indexed_fields:
            queryparser.add_prefix(field['name'], field['prefix'])

        query = queryparser.parse_query(
                querystring,
                self.PARSER_FLAGS)

        self._enquire = xapian.Enquire(cache)
        self._enquire.set_query(query)

    def get_search_results(self, offset=0, pagesize=12):
        if not self._enquire:
            return
        for match in self._enquire.get_mset(offset, pagesize):
            yield self._match_to_entry(match)

    def get_all_search_results(self):
        if not self._enquire:
            return
        offset = 0
        pagesize = 100
        while True:
            mset = self._enquire.get_mset(offset, pagesize)
            if mset.size() == 0:
                break
            for match in mset:
                yield self._match_to_entry(match)
            offset += pagesize

    def get_all_documents(self):
        cache = self._get_cache()
        postlist = cache.postlist("")
        for post in postlist:
            doc = cache.get_document(post.docid)
            yield self._doc_to_entry(doc)

    def _match_to_entry(self, match):
        return self._doc_to_entry(match.document)

    def _doc_to_entry(self, doc):
        fname = doc.get_value(0).decode('utf-8')
        if self.safe:
            data = ruamel.yaml.safe_load(doc.get_data())
        else:
            data = ruamel.yaml.load(
                doc.get_data(),
                ruamel.yaml.RoundTripLoader
                )
        entry = self.entry_class(
            self,
            data=data,
            fname=fname,
            )
        return entry

    def entry_from_uid(self, uid):
        cache = self._get_cache()
        postlist = cache.postlist('Q'+uid)
        for pitem in postlist:
            return self._doc_to_entry(cache.get_document(pitem.docid))
        return None

    def entries_from_short_uid(self, suid):
        # It would be better to search for partial UIDs inside xapian,
        # but I still can't find a way to do it, so this is a workable
        # workaround on repos where the uids are stored in the
        # filenames.
        potential_uids = [
            os.path.splitext(f)[0]
            for f in os.listdir(self.itemdir)
            if f.startswith(suid)
            ]
        return [self.entry_from_uid(u) for u in potential_uids if u]

    def remove_entries(self, uids):
        cache = xapian.WritableDatabase(
            os.path.join(self.basedir, '.lesana/xapian'),
            xapian.DB_CREATE_OR_OPEN
            )
        for uid in uids:
            for entry in self.entries_from_short_uid(uid):
                if entry is not None:
                    cache.delete_document(entry.idterm)
                    self.remove_file(entry.fname)
                else:
                    logging.warning("No such entry: {}, ignoring".format(uid))
        cache.commit()
        cache.close()

    def remove_file(self, fname):
        f_path = os.path.join(self.itemdir, fname)
        if git_available and self.settings.get('git', False):
            try:
                repo = git.Repo(self.basedir, search_parent_directories=True)
            except git.exc.InvalidGitRepositoryError as e:
                logging.warning(
                    "Could not find a git repository in {}".format(
                        self.basedir))
                return False
            repo.index.remove([f_path])
        os.remove(f_path)

    def get_template(self, template_fname, searchpath='.'):
        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                searchpath=searchpath,
                followlinks=True,
                ),
            # TODO: add autoescaping settings
            )
        try:
            template = env.get_template(template_fname)
        except jinja2.exceptions.TemplateNotFound as e:
            raise TemplatingError('Could not find template' + str(e))
        return template

    @classmethod
    def init(
            cls,
            directory=None,
            git_enabled=True,
            edit_file=None,
            settings={}
            ):
        """
        Initialize a lesana repository

        directory defaults to .
        if git_enabled is True, git support is enabled and if possible a git
        repository is initalized.
        edit_file is a syncronous function that runs on a filename
        (possibly opening the file in an editor) and should manage its
        own errors.
        """
        c_dir = os.path.abspath(directory or '.')
        os.makedirs(c_dir, exist_ok=True)
        if git_enabled:
            # Try to initalize a git repo
            if git_available:
                repo = git.Repo.init(c_dir, bare=False)
            else:
                logging.warning(
                    "python3-git not available, could not initalise " +
                    "the git repository.")
                repo = None
            # Add .lesana directory to .gitignore and add it to the
            # staging
            lesana_ignored = False
            try:
                with open(os.path.join(c_dir, '.gitignore'), 'r') as fp:
                    for line in fp:
                        if '.lesana' in line:
                            lesana_ignored = True
                            continue
            except FileNotFoundError:
                pass
            if not lesana_ignored:
                with open(os.path.join(c_dir, '.gitignore'), 'a') as fp:
                    fp.write('#Added by lesana init\n.lesana')
                if repo:
                    repo.index.add(['.gitignore'])
            # TODO: Add hook to index files as they are pulled
        # If it doesn't exist, create a skeleton of settings.yaml file
        # then open settings.yaml for editing
        filepath = os.path.join(c_dir, 'settings.yaml')
        if not os.path.exists(filepath):
            skel = resource_string(
                'lesana', 'data/settings.yaml'
                ).decode('utf-8')
            skel_dict = ruamel.yaml.load(skel, ruamel.yaml.RoundTripLoader)
            skel_dict['git'] = git_enabled
            skel_dict.update(settings)
            with open(filepath, 'w') as fp:
                ruamel.yaml.dump(
                    skel_dict,
                    stream=fp,
                    Dumper=ruamel.yaml.RoundTripDumper
                    )
        if edit_file:
            edit_file(filepath)
        if git_enabled and repo:
            repo.index.add(['settings.yaml'])
        coll = cls(c_dir)
        os.makedirs(os.path.join(coll.basedir, coll.itemdir), exist_ok=True)
        return coll


class TemplatingError(Exception):
    """
    Raised when there are errors rendering a jinja template
    """



import logging
import os.path
import shutil
import tempfile
import unittest

import git
import ruamel.yaml

import lesana


class testCollection(unittest.TestCase):
    def tearDown(self):
        shutil.rmtree(os.path.join(self.collection.basedir, '.lesana'))

    def test_empty(self):
        self.collection = lesana.Collection('tests/data/empty')
        self.assertEqual(self.collection.settings, {})

        self.collection.update_cache()
        self.assertIsNotNone(self.collection.stemmer)

    def test_load_simple(self):
        self.collection = lesana.Collection('tests/data/simple')
        self.assertIsNotNone(self.collection.settings)
        self.assertEqual(
            self.collection.settings['name'],
            "Simple lesana collection"
            )
        self.assertEqual(len(self.collection.settings['fields']), 5)
        self.assertEqual(len(self.collection.indexed_fields), 2)

        self.collection.update_cache()
        self.assertIsNotNone(self.collection.stemmer)

    def test_load_wrong_language(self):
        # This loads a collection with an invalid value in lang
        with self.assertLogs(level=logging.WARNING) as cm:
            self.collection = lesana.Collection('tests/data/wrong')
        self.assertEqual(len(cm.output), 1)
        self.assertIn("Invalid language", cm.output[0])
        # The collection will default to english, but should still work.
        self.collection.update_cache()
        self.assertIsNotNone(self.collection.settings)
        self.assertIsNotNone(self.collection.stemmer)

    def test_load_no_index_for_one_entry(self):
        # This loads a collection where some of the entries have no
        # "index" field
        with self.assertLogs(level=logging.WARNING):
            self.collection = lesana.Collection('tests/data/wrong')
        self.collection.update_cache()
        self.assertIsNotNone(self.collection.settings)
        self.assertIsNotNone(self.collection.stemmer)
        # Fields with no "index" entry are not indexed
        self.assertEqual(len(self.collection.settings['fields']), 4)
        self.assertEqual(len(self.collection.indexed_fields), 1)

    def test_load_safe(self):
        self.collection = lesana.Collection('tests/data/simple')
        self.collection.safe = True
        self.collection.update_cache()

    def test_full_search(self):
        self.collection = lesana.Collection('tests/data/simple')
        self.collection.start_search('Item')
        res = self.collection.get_all_search_results()
        matches = list(res)
        self.assertEqual(len(matches), 2)
        for m in matches:
            self.assertIsInstance(m, lesana.Entry)

    def test_search(self):
        self.collection = lesana.Collection('tests/data/simple')
        self.collection.start_search('Item')
        res = self.collection.get_search_results()
        matches = list(res)
        self.assertEqual(len(matches), 2)
        for m in matches:
            self.assertIsInstance(m, lesana.Entry)

    def test_search_wildcard(self):
        self.collection = lesana.Collection('tests/data/simple')
        self.collection.start_search('Ite*')
        res = self.collection.get_search_results()
        matches = list(res)
        self.assertEqual(len(matches), 2)
        for m in matches:
            self.assertIsInstance(m, lesana.Entry)

    def test_search_non_init(self):
        self.collection = lesana.Collection('tests/data/simple')
        matches = list(self.collection.get_search_results())
        self.assertEqual(matches, [])
        matches = list(self.collection.get_all_search_results())
        self.assertEqual(matches, [])

    def test_all_entries(self):
        self.collection = lesana.Collection('tests/data/simple')
        res = self.collection.get_all_documents()
        matches = list(res)
        self.assertEqual(len(matches), 3)
        for m in matches:
            self.assertIsInstance(m, lesana.Entry)

    def test_entry_from_uid(self):
        self.collection = lesana.Collection('tests/data/simple')
        entry = self.collection.entry_from_uid(
            '11189ee47ddf4796b718a483b379f976'
            )
        self.assertEqual(entry.uid, '11189ee47ddf4796b718a483b379f976')
        self.collection.safe = True
        entry = self.collection.entry_from_uid(
            '11189ee47ddf4796b718a483b379f976'
            )
        self.assertEqual(entry.uid, '11189ee47ddf4796b718a483b379f976')

    def test_entry_from_short_uid(self):
        self.collection = lesana.Collection('tests/data/simple')
        entries = self.collection.entries_from_short_uid(
            '11189ee4'
            )
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0].uid, '11189ee47ddf4796b718a483b379f976')
        entries = self.collection.entries_from_short_uid(
            '11189ee47ddf4796b718a483b379f976'
            )
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0].uid, '11189ee47ddf4796b718a483b379f976')
        entries = self.collection.entries_from_short_uid(
            '12345678'
            )
        self.assertEqual(len(entries), 0)

    def test_index_missing_file(self):
        self.collection = lesana.Collection('tests/data/simple')
        with self.assertLogs(level=logging.WARNING) as cm:
            self.collection.update_cache(['non_existing_file'])
        self.assertEqual(len(cm.output), 1)
        self.assertIn("non_existing_file", cm.output[0])

    def test_get_entry_missing_uid(self):
        self.collection = lesana.Collection('tests/data/simple')
        entry = self.collection.entry_from_uid('this is not an uid')
        self.assertIsNone(entry)

    def test_render_collection(self):
        self.collection = lesana.Collection('tests/data/simple')
        template = self.collection.get_template(
            'tests/data/simple/templates/collection_template.txt'
            )
        res = template.render(entries=self.collection.get_all_documents())
        self.assertIn('11189ee4: Another item', res)


class testEntries(unittest.TestCase):
    def setUp(self):
        self.collection = lesana.Collection('tests/data/simple')
        self.basepath = 'tests/data/simple/items'
        self.filenames = []

    def tearDown(self):
        for fname in self.filenames:
            os.remove(fname)
        shutil.rmtree(os.path.join(self.collection.basedir, '.lesana'))

    def test_simple(self):
        fname = '085682ed-6792-499d-a3ab-9aebd683c011.yaml'
        with open(os.path.join(self.basepath, fname)) as fp:
            data = ruamel.yaml.safe_load(fp)
        entry = lesana.Entry(self.collection, data=data, fname=fname)
        self.assertEqual(entry.idterm, 'Q'+data['uid'])
        fname = '11189ee47ddf4796b718a483b379f976.yaml'
        uid = '11189ee47ddf4796b718a483b379f976'
        with open(os.path.join(self.basepath, fname)) as fp:
            data = ruamel.yaml.safe_load(fp)
        entry = lesana.Entry(self.collection, data=data, fname=fname)
        self.assertEqual(entry.idterm, 'Q'+uid)
        self.assertEqual(entry.short_id, uid[:8])

    def test_write_new(self):
        new_entry = lesana.Entry(self.collection)
        self.collection.save_entries(entries=[new_entry])
        entry_fname = 'tests/data/simple/items/' + new_entry.fname
        self.filenames.append(entry_fname)
        with open(entry_fname) as fp:
            text = fp.read()
        self.assertIn('quantity (integer): how many items are there', text)
        self.assertIn('other (yaml):', text)
        self.assertNotIn('position (string)', text)
        written = ruamel.yaml.safe_load(text)
        self.assertIsInstance(written['quantity'], int)
        self.assertIsInstance(written['name'], str)

    def test_entry_representation(self):
        uid = '11189ee47ddf4796b718a483b379f976'
        entry = self.collection.entry_from_uid(uid)
        self.assertEqual(
            str(entry),
            uid
            )
        label = '{{ uid }}: {{ name }}'
        self.collection.settings['entry_label'] = label
        self.assertEqual(
            str(entry),
            '{uid}: {name}'.format(uid=uid, name='Another item')
            )

    def test_entry_creation_uid_but_no_filename(self):
        fname = '11189ee47ddf4796b718a483b379f976.yaml'
        with open(os.path.join(self.basepath, fname)) as fp:
            data = ruamel.yaml.safe_load(fp)
        data['uid'] = '11189ee47ddf4796b718a483b379f976'
        entry = lesana.Entry(self.collection, data=data)
        self.assertEqual(entry.fname, fname)

    def test_entry_creation_no_uid_no_filename(self):
        fname = '11189ee47ddf4796b718a483b379f976.yaml'
        with open(os.path.join(self.basepath, fname)) as fp:
            data = ruamel.yaml.safe_load(fp)
        entry = lesana.Entry(self.collection, data=data)
        self.assertIsNotNone(entry.uid)
        self.assertIsNotNone(entry.fname)

    def test_entry_creation_filename_but_no_uid(self):
        fname = '11189ee47ddf4796b718a483b379f976.yaml'
        uid = '11189ee47ddf4796b718a483b379f976'
        with open(os.path.join(self.basepath, fname)) as fp:
            data = ruamel.yaml.safe_load(fp)
        entry = lesana.Entry(self.collection, data=data, fname=fname)
        self.assertEqual(entry.uid, uid)

    def test_entry_str_filename_and_uid(self):
        fname = '11189ee47ddf4796b718a483b379f976.yaml'
        with open(os.path.join(self.basepath, fname)) as fp:
            data = ruamel.yaml.safe_load(fp)
        data['uid'] = '11189ee47ddf4796b718a483b379f976'
        entry = lesana.Entry(self.collection, data=data)
        self.assertEqual(str(entry), data['uid'])
        self.collection.settings['entry_label'] = '{{ uid }}: {{ name }}'
        self.assertEqual(str(entry), data['uid'] + ': Another item')

    def test_entry_str_filename_no_uid(self):
        fname = '11189ee47ddf4796b718a483b379f976.yaml'
        with open(os.path.join(self.basepath, fname)) as fp:
            data = ruamel.yaml.safe_load(fp)
        entry = lesana.Entry(self.collection, data=data)
        uid = entry.uid
        self.assertEqual(str(entry), uid)
        self.collection.settings['entry_label'] = '{{ uid }}: {{ name }}'
        self.assertEqual(str(entry), uid + ': Another item')

    def test_render_entry(self):
        fname = '11189ee47ddf4796b718a483b379f976.yaml'
        with open(os.path.join(self.basepath, fname)) as fp:
            data = ruamel.yaml.safe_load(fp)
        entry = lesana.Entry(self.collection, data=data)
        uid = entry.uid
        res = entry.render('tests/data/simple/templates/trivial_template.txt')
        self.assertIn(uid, res)

    def test_empty_data(self):
        entry = lesana.Entry(self.collection)
        self.assertIn("name: ''", entry.yaml_data)
        self.assertIn('quantity: 0', entry.yaml_data)


class testComplexCollection(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.collection = lesana.Collection('tests/data/complex')

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(os.path.join(self.collection.basedir, '.lesana'))

    def test_init(self):
        self.assertIsNotNone(self.collection.settings)
        self.assertEqual(
            self.collection.settings['name'],
            "Fully featured lesana collection"
            )
        self.assertEqual(len(self.collection.settings['fields']), 9)
        self.assertIsNotNone(self.collection.stemmer)
        self.assertEqual(len(self.collection.indexed_fields), 6)

    def test_index(self):
        self.collection.update_cache()

    def test_indexing_list(self):
        self.collection.update_cache(['73097121f1874a6ea2f927db7dc4f11e.yaml'])
        self.collection.start_search('tags:this')
        res = self.collection.get_search_results()
        matches = list(res)
        self.assertEqual(len(matches), 1)
        for m in matches:
            self.assertIsInstance(m, lesana.Entry)

    def test_boolean_field(self):
        entry = self.collection.entry_from_uid(
            '73097121f1874a6ea2f927db7dc4f11e'
            )
        self.assertIsInstance(entry.data['exists'], bool)

    def test_empty_data(self):
        entry = lesana.Entry(self.collection)
        print(entry.yaml_data)
        self.assertIn("name: ''", entry.yaml_data)
        self.assertIn('with_default', entry.yaml_data)
        self.assertIn('amount: 0', entry.yaml_data)


class testCollectionWithErrors(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.collection = lesana.Collection('tests/data/wrong')

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(os.path.join(self.collection.basedir, '.lesana'))

    def test_init(self):
        self.assertIsNotNone(self.collection.settings)
        self.assertEqual(
            self.collection.settings['name'],
            "Lesana collection with certain errors"
            )
        self.assertEqual(len(self.collection.settings['fields']), 4)
        self.assertIsNotNone(self.collection.stemmer)
        self.assertEqual(len(self.collection.indexed_fields), 1)

    def test_index(self):
        loaded = self.collection.update_cache()
        self.assertEqual(loaded, 0)


class testCollectionCreation(unittest.TestCase):
    def test_init(self):
        tmpdir = tempfile.mkdtemp()
        collection = lesana.Collection.init(tmpdir)
        self.assertIsInstance(collection, lesana.Collection)
        self.assertTrue(os.path.isdir(os.path.join(tmpdir, '.git')))
        self.assertTrue(os.path.isdir(os.path.join(tmpdir, '.lesana')))
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, 'settings.yaml')))
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, '.gitignore')))
        # and then run it twice on the same directory, nothing should break
        collection = lesana.Collection.init(tmpdir)
        self.assertIsInstance(collection, lesana.Collection)
        self.assertTrue(os.path.isdir(os.path.join(tmpdir, '.git')))
        self.assertTrue(os.path.isdir(os.path.join(tmpdir, '.lesana')))
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, 'settings.yaml')))
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, '.gitignore')))
        created = lesana.Collection(tmpdir)
        self.assertTrue(created.settings['git'])
        shutil.rmtree(tmpdir)

    def do_nothing(*args, **kwargs):
        # A function that does nothing instead of editing a file
        pass

    def test_init_edit_file(self):
        tmpdir = tempfile.mkdtemp()
        collection = lesana.Collection.init(tmpdir, edit_file=self.do_nothing)
        self.assertIsInstance(collection, lesana.Collection)
        self.assertTrue(os.path.isdir(os.path.join(tmpdir, '.git')))
        self.assertTrue(os.path.isdir(os.path.join(tmpdir, '.lesana')))
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, 'settings.yaml')))
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, '.gitignore')))
        shutil.rmtree(tmpdir)

    def test_init_no_git(self):
        tmpdir = tempfile.mkdtemp()
        collection = lesana.Collection.init(tmpdir, git_enabled=False)
        self.assertIsInstance(collection, lesana.Collection)
        self.assertFalse(os.path.isdir(os.path.join(tmpdir, '.git')))
        self.assertTrue(os.path.isdir(os.path.join(tmpdir, '.lesana')))
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, 'settings.yaml')))
        self.assertFalse(os.path.isfile(os.path.join(tmpdir, '.gitignore')))
        # and then run it twice on the same directory, nothing should break
        collection = lesana.Collection.init(tmpdir, git_enabled=False)
        self.assertIsInstance(collection, lesana.Collection)
        self.assertFalse(os.path.isdir(os.path.join(tmpdir, '.git')))
        self.assertTrue(os.path.isdir(os.path.join(tmpdir, '.lesana')))
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, 'settings.yaml')))
        self.assertFalse(os.path.isfile(os.path.join(tmpdir, '.gitignore')))
        created = lesana.Collection(tmpdir)
        self.assertFalse(created.settings['git'])
        shutil.rmtree(tmpdir)

    def test_deletion(self):
        tmpdir = tempfile.mkdtemp()
        shutil.copy('tests/data/simple/settings.yaml', tmpdir)
        shutil.copytree(
            'tests/data/simple/items',
            os.path.join(tmpdir, 'items'),
            )
        collection = lesana.Collection.init(tmpdir)
        # We start with one item indexed with the term "another"
        collection.start_search('another')
        mset = collection._enquire.get_mset(0, 10)
        self.assertEqual(mset.get_matches_estimated(), 1)
        # Then delete it
        collection.remove_entries(['11189ee47ddf4796b718a483b379f976'])
        # An now we should have none
        self.assertFalse(os.path.exists(os.path.join(
            tmpdir,
            'items',
            '11189ee47ddf4796b718a483b379f976.yaml'
            )))
        collection.start_search('another')
        mset = collection._enquire.get_mset(0, 10)
        self.assertEqual(mset.get_matches_estimated(), 0)

    def test_partial_uid_deletion(self):
        tmpdir = tempfile.mkdtemp()
        shutil.copy('tests/data/simple/settings.yaml', tmpdir)
        shutil.copytree(
            'tests/data/simple/items',
            os.path.join(tmpdir, 'items'),
            )
        collection = lesana.Collection.init(tmpdir)
        # We start with one item indexed with the term "another"
        collection.start_search('another')
        mset = collection._enquire.get_mset(0, 10)
        self.assertEqual(mset.get_matches_estimated(), 1)
        # Then delete it, using the short id
        collection.remove_entries(['11189ee4'])
        # An now we should have none
        self.assertFalse(os.path.exists(os.path.join(
            tmpdir,
            'items',
            '11189ee47ddf4796b718a483b379f976.yaml'
            )))
        collection.start_search('another')
        mset = collection._enquire.get_mset(0, 10)
        self.assertEqual(mset.get_matches_estimated(), 0)

    def _find_file_in_git_index(self, fname, index):
        found = False
        for (path, stage) in index.entries:
            if fname in path:
                found = True
                break
        return found

    def test_git_adding(self):
        tmpdir = tempfile.mkdtemp()
        shutil.copy('tests/data/simple/settings.yaml', tmpdir)
        shutil.copytree(
            'tests/data/simple/items',
            os.path.join(tmpdir, 'items'),
            )
        collection = lesana.Collection.init(tmpdir)
        fname = '11189ee47ddf4796b718a483b379f976.yaml'
        repo = git.Repo(tmpdir)
        # By default, this collection doesn't have any git entry in the
        # settings (but there is a repo)
        collection.git_add_files([os.path.join(collection.itemdir, fname)])
        self.assertFalse(self._find_file_in_git_index(fname, repo.index))
        # Then we set it to false
        collection.settings['git'] = False
        collection.git_add_files([os.path.join(collection.itemdir, fname)])
        self.assertFalse(self._find_file_in_git_index(fname, repo.index))
        # And only when it's set to true we should find the file in the
        # staging area
        collection.settings['git'] = True
        collection.git_add_files([os.path.join(collection.itemdir, fname)])
        self.assertTrue(self._find_file_in_git_index(fname, repo.index))
        shutil.rmtree(tmpdir)

    def test_init_custom_settings(self):
        tmpdir = tempfile.mkdtemp()
        collection = lesana.Collection.init(
            tmpdir,
            edit_file=self.do_nothing,
            settings={
                'name': 'A different name',
                'fields': [
                    {'name': 'title', 'type': 'string'},
                    {'name': 'author', 'type': 'string'},
                    ],
                },
            )
        self.assertIsInstance(collection, lesana.Collection)
        self.assertTrue(os.path.isfile(os.path.join(tmpdir, 'settings.yaml')))
        self.assertEqual(collection.settings['name'], 'A different name')
        self.assertEqual(len(collection.settings['fields']), 2)
        shutil.rmtree(tmpdir)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

try:
    with open("README.rst", 'r') as fp:
        long_description = fp.read()
except IOError:
    print("Could not read README.rst, long_description will be empty.")
    long_description = ""

setup(
    name='lesana',
    version='0.4.1',
    packages=find_packages(),
    scripts=['scripts/lesana'],

    package_data={
        '': ['*.yaml']
        },
    test_suite='tests',

    install_requires=[
        # 'xapian >= 1.4',
        'ruamel.yaml',
        'jinja2',
        ],
    python_requires='>=3',

    # Metadata
    author="Elena ``of Valhalla'' Grandi",
    author_email='valhalla@trueelena.org',
    description='Manage collection inventories throught yaml files.',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    license='GPLv3+',
    keywords='collection inventory',
    url='https://lesana.trueelena.org/lesana',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        ],
    project_urls={
        'Source': 'https://git.trueelena.org/cgit.cgi/software/lesana/',
        'Documentation': 'https://git.trueelena.org/cgit.cgi/software/lesana/tree/docs',
        }
)
